#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This file is part of PLCOpenEditor, a library implementing an IEC 61131-3 editor
#based on the plcopen standard. 
#
#Copyright (C) 2007: Edouard TISSERANT and Laurent BESSARD
#
#See COPYING file for copyrights details.
#
#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.
#
#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#General Public License for more details.
#
#You should have received a copy of the GNU General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import wx

from plcopen.structures import *
from PLCControler import ITEM_POU

from Dialogs import *

SCROLLBAR_UNIT = 10
WINDOW_BORDER = 10
SCROLL_ZONE = 10

def AppendMenu(parent, help, id, kind, text):
    if wx.VERSION >= (2, 6, 0):
        parent.Append(help=help, id=id, kind=kind, text=text)
    else:
        parent.Append(helpString=help, id=id, kind=kind, item=text)

if wx.Platform == '__WXMSW__':
    faces = { 'times': 'Times New Roman',
              'mono' : 'Courier New',
              'helv' : 'Arial',
              'other': 'Comic Sans MS',
              'size' : 20,
             }
else:
    faces = { 'times': 'Times',
              'mono' : 'Courier',
              'helv' : 'Helvetica',
              'other': 'new century schoolbook',
              'size' : 20,
             }

#-------------------------------------------------------------------------------
#                       Graphic elements Viewer base class
#-------------------------------------------------------------------------------

# ID Constants for alignment menu items
[ID_VIEWERALIGNMENTMENUITEMS0, ID_VIEWERALIGNMENTMENUITEMS1,
 ID_VIEWERALIGNMENTMENUITEMS2, ID_VIEWERALIGNMENTMENUITEMS4,
 ID_VIEWERALIGNMENTMENUITEMS5, ID_VIEWERALIGNMENTMENUITEMS6,
] = [wx.NewId() for _init_coll_AlignmentMenu_Items in range(6)]

# ID Constants for contextual menu items
[ID_VIEWERCONTEXTUALMENUITEMS0, ID_VIEWERCONTEXTUALMENUITEMS1,
 ID_VIEWERCONTEXTUALMENUITEMS2, ID_VIEWERCONTEXTUALMENUITEMS3,
 ID_VIEWERCONTEXTUALMENUITEMS5, ID_VIEWERCONTEXTUALMENUITEMS6,
 ID_VIEWERCONTEXTUALMENUITEMS8, ID_VIEWERCONTEXTUALMENUITEMS9,
 ID_VIEWERCONTEXTUALMENUITEMS11, ID_VIEWERCONTEXTUALMENUITEMS12,
 ID_VIEWERCONTEXTUALMENUITEMS14, ID_VIEWERCONTEXTUALMENUITEMS16,
 ID_VIEWERCONTEXTUALMENUITEMS17,
] = [wx.NewId() for _init_coll_ContextualMenu_Items in range(13)]


class ViewerDropTarget(wx.TextDropTarget):
    
    def __init__(self, parent):
        wx.TextDropTarget.__init__(self)
        self.ParentWindow = parent
    
    def OnDropText(self, x, y, data):
        x, y = self.ParentWindow.CalcUnscrolledPosition(x, y)
        scaling = self.ParentWindow.Scaling
        values = eval(data)
        if values[1] == "program":
            message = wx.MessageDialog(self.ParentWindow, "Programs can't be used by other POUs!", "Error", wx.OK|wx.ICON_ERROR)
            message.ShowModal()
            message.Destroy()
        elif values[1] in ["function", "functionBlock", "program"]:
            name, type = self.ParentWindow.Controler.GetEditedElementType(self.ParentWindow.GetTagName())
            if name == values[0]:
                message = wx.MessageDialog(self.ParentWindow, "\"%s\" can't use itself!"%name, "Error", wx.OK|wx.ICON_ERROR)
                message.ShowModal()
                message.Destroy()
            elif type == "function" and values[1] != "function":
                message = wx.MessageDialog(self.ParentWindow, "Function Blocks can't be used by Functions!", "Error", wx.OK|wx.ICON_ERROR)
                message.ShowModal()
                message.Destroy()
            elif self.ParentWindow.Controler.PouIsUsedBy(name, values[0]):
                message = wx.MessageDialog(self.ParentWindow, "\"%s\" is already used by \"%s\"!"%(name, values[0]), "Error", wx.OK|wx.ICON_ERROR)
                message.ShowModal()
                message.Destroy()
            else:
                blockname = values[2]
                if values[1] != "function" and blockname == "":
                    dialog = wx.TextEntryDialog(self.ParentWindow, "Block name", "Please enter a block name", "", wx.OK|wx.CANCEL|wx.CENTRE)
                    if dialog.ShowModal():
                        blockname = dialog.GetValue()
                    dialog.Destroy()
                if blockname.upper() in [name.upper() for name in self.ParentWindow.Controler.GetProjectPouNames()]:
                    message = wx.MessageDialog(self.ParentWindow, "\"%s\" pou already exists!"%blockname, "Error", wx.OK|wx.ICON_ERROR)
                    message.ShowModal()
                    message.Destroy()
                elif blockname.upper() in [name.upper() for name in self.ParentWindow.Controler.GetEditedElementVariables(self.ParentWindow.GetTagName())]:
                    message = wx.MessageDialog(self.ParentWindow, "\"%s\" element for this pou already exists!"%blockname, "Error", wx.OK|wx.ICON_ERROR)
                    message.ShowModal()
                    message.Destroy()
                else:
                    id = self.ParentWindow.GetNewId()
                    block = FBD_Block(self.ParentWindow, values[0], blockname, id)
                    width, height = block.GetMinSize()
                    if scaling is not None:
                        x = round(float(x) / float(scaling[0])) * scaling[0]
                        y = round(float(y) / float(scaling[1])) * scaling[1]
                        width = round(float(width) / float(scaling[0]) + 0.5) * scaling[0]
                        height = round(float(height) / float(scaling[1]) + 0.5) * scaling[1]
                    block.SetPosition(x, y)
                    block.SetSize(width, height)
                    self.ParentWindow.AddBlock(block)
                    self.ParentWindow.Controler.AddEditedElementBlock(self.ParentWindow.GetTagName(), id, values[0], blockname)
                    self.ParentWindow.RefreshBlockModel(block)
                    self.ParentWindow.RefreshBuffer()
                    self.ParentWindow.RefreshScrollBars()
                    self.ParentWindow.ParentWindow.RefreshVariablePanel(self.ParentWindow.GetTagName())
                    self.ParentWindow.Refresh(False)
        elif values[1] != "location":
            if values[3] == self.ParentWindow.GetTagName():
                id = self.ParentWindow.GetNewId()
                if values[1] == "Output":
                    var_type = OUTPUT
                elif values[1] == "InOut":
                    var_type = INPUT
                else:
                    var_type = INPUT
                variable = FBD_Variable(self.ParentWindow, var_type, values[0], values[2], id)
                width, height = variable.GetMinSize()
                if scaling is not None:
                    x = round(float(x) / float(scaling[0])) * scaling[0]
                    y = round(float(y) / float(scaling[1])) * scaling[1]
                    width = round(float(width) / float(scaling[0]) + 0.5) * scaling[0]
                    height = round(float(height) / float(scaling[1]) + 0.5) * scaling[1]
                variable.SetPosition(x, y)
                variable.SetSize(width, height)
                self.ParentWindow.AddBlock(variable)
                self.ParentWindow.Controler.AddEditedElementVariable(self.ParentWindow.GetTagName(), id, var_type)
                self.ParentWindow.RefreshVariableModel(variable)
                self.ParentWindow.RefreshBuffer()
                self.ParentWindow.RefreshScrollBars()
                self.ParentWindow.Refresh(False)
            else:
                message = wx.MessageDialog(self.ParentWindow, "Variable don't belong to this POU!", "Error", wx.OK|wx.ICON_ERROR)
                message.ShowModal()
                message.Destroy()
            
if wx.VERSION >= (2, 8, 0):
    import wx.aui

    class MDIViewer(wx.aui.AuiMDIChildFrame):
        def __init__(self, parent, tagname, window, controler):
            wx.aui.AuiMDIChildFrame.__init__(self, parent, -1, title = "")
            
            sizer = wx.BoxSizer(wx.HORIZONTAL)
            
            self.Viewer = Viewer(self, tagname, window, controler)
            
            sizer.AddWindow(self.Viewer, 1, border=0, flag=wx.GROW)
            
            self.SetSizer(sizer)
        
        def GetViewer(self):
            return self.Viewer

"""
Class that implements a Viewer based on a wx.ScrolledWindow for drawing and 
manipulating graphic elements
"""

class Viewer(wx.ScrolledWindow):
    
    if wx.VERSION < (2, 6, 0):
        def Bind(self, event, function, id = None):
            if id is not None:
                event(self, id, function)
            else:
                event(self, function)
    
    # Create Alignment Menu items
    def _init_coll_AlignmentMenu_Items(self, parent):
        # Create menu items
        AppendMenu(parent, help='', id=ID_VIEWERALIGNMENTMENUITEMS0,
              kind=wx.ITEM_NORMAL, text=u'Left')
        AppendMenu(parent, help='', id=ID_VIEWERALIGNMENTMENUITEMS1,
              kind=wx.ITEM_NORMAL, text=u'Center')
        AppendMenu(parent, help='', id=ID_VIEWERALIGNMENTMENUITEMS2,
              kind=wx.ITEM_NORMAL, text=u'Right')
        parent.AppendSeparator()
        AppendMenu(parent, help='', id=ID_VIEWERALIGNMENTMENUITEMS4,
              kind=wx.ITEM_NORMAL, text=u'Top')
        AppendMenu(parent, help='', id=ID_VIEWERALIGNMENTMENUITEMS5,
              kind=wx.ITEM_NORMAL, text=u'Middle')
        AppendMenu(parent, help='', id=ID_VIEWERALIGNMENTMENUITEMS6,
              kind=wx.ITEM_NORMAL, text=u'Bottom')
        # Link menu event to corresponding called functions
        self.Bind(wx.EVT_MENU, self.OnAlignLeftMenu,
              id=ID_VIEWERALIGNMENTMENUITEMS0)
        self.Bind(wx.EVT_MENU, self.OnAlignCenterMenu,
              id=ID_VIEWERALIGNMENTMENUITEMS1)
        self.Bind(wx.EVT_MENU, self.OnAlignRightMenu,
              id=ID_VIEWERALIGNMENTMENUITEMS2)
        self.Bind(wx.EVT_MENU, self.OnAlignTopMenu,
              id=ID_VIEWERALIGNMENTMENUITEMS4)
        self.Bind(wx.EVT_MENU, self.OnAlignMiddleMenu,
              id=ID_VIEWERALIGNMENTMENUITEMS5)
        self.Bind(wx.EVT_MENU, self.OnAlignBottomMenu,
              id=ID_VIEWERALIGNMENTMENUITEMS6)
    
    # Create Contextual Menu items
    def _init_coll_ContextualMenu_Items(self, parent):
        # Create menu items
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS0,
              kind=wx.ITEM_RADIO, text=u'No Modifier')
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS1,
              kind=wx.ITEM_RADIO, text=u'Negated')
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS2,
              kind=wx.ITEM_RADIO, text=u'Rising Edge')
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS3,
              kind=wx.ITEM_RADIO, text=u'Falling Edge')
        parent.AppendSeparator()
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS5,
              kind=wx.ITEM_NORMAL, text=u'Add Wire Segment')
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS6,
              kind=wx.ITEM_NORMAL, text=u'Delete Wire Segment')
        parent.AppendSeparator()
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS8,
              kind=wx.ITEM_NORMAL, text=u'Add Divergence Branch')
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS9,
              kind=wx.ITEM_NORMAL, text=u'Delete Divergence Branch')
        parent.AppendSeparator()
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS11,
              kind=wx.ITEM_NORMAL, text=u'Clear Execution Order')
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS12,
              kind=wx.ITEM_NORMAL, text=u'Reset Execution Order')
        parent.AppendSeparator()
        parent.AppendMenu(ID_VIEWERCONTEXTUALMENUITEMS14, "Alignment", self.AlignmentMenu)
        parent.AppendSeparator()
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS16,
              kind=wx.ITEM_NORMAL, text=u'Edit Block')
        AppendMenu(parent, help='', id=ID_VIEWERCONTEXTUALMENUITEMS17,
              kind=wx.ITEM_NORMAL, text=u'Delete')
        # Link menu event to corresponding called functions
        self.Bind(wx.EVT_MENU, self.OnNoModifierMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS0)
        self.Bind(wx.EVT_MENU, self.OnNegatedMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS1)
        self.Bind(wx.EVT_MENU, self.OnRisingEdgeMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS2)
        self.Bind(wx.EVT_MENU, self.OnFallingEdgeMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS3)
        self.Bind(wx.EVT_MENU, self.OnAddSegmentMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS5)
        self.Bind(wx.EVT_MENU, self.OnDeleteSegmentMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS6)
        self.Bind(wx.EVT_MENU, self.OnAddBranchMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS8)
        self.Bind(wx.EVT_MENU, self.OnDeleteBranchMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS9)
        self.Bind(wx.EVT_MENU, self.OnClearExecutionOrderMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS11)
        self.Bind(wx.EVT_MENU, self.OnResetExecutionOrderMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS12)
        self.Bind(wx.EVT_MENU, self.OnEditBlockMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS16)
        self.Bind(wx.EVT_MENU, self.OnDeleteMenu,
              id=ID_VIEWERCONTEXTUALMENUITEMS17)
    
    # Create and initialize Contextual Menu
    def _init_menus(self):
        self.AlignmentMenu = wx.Menu(title='')
        self.ContextualMenu = wx.Menu(title='')
        
        self._init_coll_AlignmentMenu_Items(self.AlignmentMenu)
        self._init_coll_ContextualMenu_Items(self.ContextualMenu)
    
    # Create a new Viewer
    def __init__(self, parent, tagname, window, controler):
        wx.ScrolledWindow.__init__(self, parent, pos=wx.Point(0, 0), size=wx.Size(0, 0), 
            style=wx.HSCROLL | wx.VSCROLL)
        self._init_menus()
        # Adding a rubberband to Viewer
        self.rubberBand = RubberBand(drawingSurface=self)
        self.SetBackgroundColour(wx.Colour(255,255,255))
        self.ResetView()
        self.Scaling = None
        self.DrawGrid = True
        self.GridBrush = wx.TRANSPARENT_BRUSH
        self.DrawingWire = False
        self.current_id = 0
        self.TagName = tagname
        
        # Initialize Block, Wire and Comment numbers
        self.block_id = self.wire_id = self.comment_id = 0
        
        # Initialize Viewer mode to Selection mode
        self.Mode = MODE_SELECTION
        self.SavedMode = False
        self.CurrentLanguage = "FBD"
        
        self.ParentWindow = window
        self.Controler = controler
        
        self.SetDropTarget(ViewerDropTarget(self))
        
        dc = wx.ClientDC(self)
        font = wx.Font(faces["size"], wx.DEFAULT, wx.NORMAL, wx.NORMAL, faceName = faces["mono"])
        dc.SetFont(font)
        width, height = dc.GetTextExtent("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
        while width > 260:
            faces["size"] -= 1
            font = wx.Font(faces["size"], wx.DEFAULT, wx.NORMAL, wx.NORMAL, faceName = faces["mono"])
            dc.SetFont(font)
            width, height = dc.GetTextExtent("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
        self.SetFont(font)
        
        # Link Viewer event to corresponding methods
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnViewerLeftDown)
        self.Bind(wx.EVT_LEFT_UP, self.OnViewerLeftUp)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnViewerLeftDClick)
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnViewerRightDown)
        self.Bind(wx.EVT_RIGHT_UP, self.OnViewerRightUp)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnLeaveViewer)
        self.Bind(wx.EVT_MOTION, self.OnViewerMotion)
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_SCROLLWIN, self.OnMoveWindow)
        self.Bind(wx.EVT_SIZE, self.OnMoveWindow)
    
    def GetScrolledRect(self, rect):
        rect.x, rect.y = self.CalcScrolledPosition(rect.x, rect.y)
        return rect
    
    def GetScaling(self):
        return self.Scaling
    
    def SetTagName(self, tagname):
        self.TagName = tagname
        
    def GetTagName(self):
        return self.TagName
    
    def IsViewing(self, tagname):
        return self.TagName == tagname
    
    # Returns a new id
    def GetNewId(self):
        self.current_id += 1
        return self.current_id
    
    # Destructor
    def __del__(self):
        self.ResetView()

    def GetLogicalDC(self, buffered=False):
        if buffered:
            dc = wx.BufferedPaintDC(self)
        else:
            dc = wx.ClientDC(self)
        dc.SetFont(self.GetFont())
        if wx.VERSION >= (2, 6, 0):
            self.DoPrepareDC(dc)
        else:
            self.PrepareDC(dc)
        return dc

#-------------------------------------------------------------------------------
#                         Element management functions
#-------------------------------------------------------------------------------

    def AddBlock(self, block):
        self.block_id += 1
        self.Blocks[block] = self.block_id
        
    def AddWire(self, wire):
        self.wire_id += 1
        self.Wires[wire] = self.wire_id
        
    def AddComment(self, comment):
        self.comment_id += 1
        self.Comments[comment] = self.comment_id

    def IsBlock(self, block):
        return self.Blocks.get(block, False)
        
    def IsWire(self, wire):
        return self.Wires.get(wire, False)
        
    def IsComment(self, comment):
        return self.Comments.get(comment, False)

    def RemoveBlock(self, block):
        self.Blocks.pop(block)
        
    def RemoveWire(self, wire):
        self.Wires.pop(wire)
        
    def RemoveComment(self, comment):
        self.Comments.pop(comment)

    def GetElements(self, sort_blocks=False, sort_wires=False, sort_comments=False):
        blocks = self.Blocks.keys()
        wires = self.Wires.keys()
        comments = self.Comments.keys()
        if sort_blocks:
            blocks.sort(lambda x,y:self.Blocks[x].__cmp__(self.Blocks[y]))
        if sort_wires:
            wires.sort(lambda x,y:self.Wires[x].__cmp__(self.Wires[y]))
        if sort_comments:
            comments.sort(lambda x,y:self.Comments[x].__cmp__(self.Comments[y]))
        return blocks + wires + comments

#-------------------------------------------------------------------------------
#                              Reset functions
#-------------------------------------------------------------------------------

    # Resets Viewer lists
    def ResetView(self):
        self.Blocks = {}
        self.Wires = {}
        self.Comments = {}
        self.SelectedElement = None
        self.HighlightedElement = None
    
    # Remove all elements
    def CleanView(self):
        for block in self.Blocks.keys():
            block.Clean()
        self.ResetView()
    
    # Changes Viewer mode
    def SetMode(self, mode):
        if self.Mode != mode or mode == MODE_SELECTION:
            self.Mode = mode
            self.SavedMode = False
        else:
            self.SavedMode = True
        # Reset selection
        if self.Mode != MODE_SELECTION and self.SelectedElement:
            self.SelectedElement.SetSelected(False)
            self.SelectedElement = None
    
    # Return current drawing mode
    def GetDrawingMode(self):
        return self.ParentWindow.GetDrawingMode()
    
    # Buffer the last model state
    def RefreshBuffer(self):
        self.Controler.BufferProject()
        self.ParentWindow.RefreshTitle()
        self.ParentWindow.RefreshEditMenu()

    # Refresh the current scaling
    def RefreshScaling(self, refresh=True):
        properties = self.Controler.GetProjectProperties()
        scaling = properties["scaling"][self.CurrentLanguage]
        if scaling != (0, 0):
            self.Scaling = scaling
            if self.DrawGrid:
                bitmap = wx.EmptyBitmap(*scaling)
                dc = wx.MemoryDC(bitmap)
                dc.SetBackground(wx.Brush(self.GetBackgroundColour()))
                dc.Clear()
                dc.SetPen(wx.Pen(wx.Colour(180, 180, 180)))
                dc.DrawPoint(0, 0)
                self.GridBrush = wx.BrushFromBitmap(bitmap)
            else:
                self.GridBrush = wx.TRANSPARENT_BRUSH
        else:
            self.Scaling = None
            self.GridBrush = wx.TRANSPARENT_BRUSH
        if refresh:
            self.Refresh(False)
        
        
#-------------------------------------------------------------------------------
#                          Refresh functions
#-------------------------------------------------------------------------------

    def ResetBuffer(self):
        pass

    # Refresh Viewer elements
    def RefreshView(self):
        self.current_id = 0
        # Start by reseting Viewer
        self.ResetView()
        instance = True
        # List of ids of already loaded blocks
        ids = []
        # Load Blocks until they are all loaded
        while instance:
            instance = self.Controler.GetEditedElementInstanceInfos(self.TagName, exclude=ids)
            if instance:
                self.loadInstance(instance, ids)
        self.RefreshScrollBars()
        
        to_delete = []
        for wire in self.Wires:
            if not wire.IsConnectedCompatible():
                to_delete.append(wire)
        for wire in to_delete:
            wire.Delete()
    
        self.Refresh(False)
    
    def RefreshScrollBars(self):
        xstart, ystart = self.GetViewStart()
        window_size = self.GetClientSize()
        maxx = maxy = 0
        for element in self.GetElements():
            bbox = element.GetBoundingBox()
            maxx = max(maxx, bbox.x + bbox.width)
            maxy = max(maxy, bbox.y + bbox.height)
        maxx = max(maxx + WINDOW_BORDER, xstart * SCROLLBAR_UNIT + window_size[0])
        maxy = max(maxy + WINDOW_BORDER, ystart * SCROLLBAR_UNIT + window_size[1])
        if self.rubberBand.IsShown():
            extent = self.rubberBand.GetCurrentExtent()
            maxx = max(maxx, extent.x + extent.width)
            maxy = max(maxy, extent.y + extent.height)
        self.SetScrollbars(SCROLLBAR_UNIT, SCROLLBAR_UNIT, 
            maxx / SCROLLBAR_UNIT, maxy / SCROLLBAR_UNIT, xstart, ystart, True)
        
    # Load instance from given informations
    def loadInstance(self, instance, ids):
        ids.append(instance["id"])
        self.current_id = max(self.current_id, instance["id"]) 
        if instance["type"] == "input":
            variable = FBD_Variable(self, INPUT, instance["name"], instance["value_type"], instance["id"], instance["executionOrder"])
            variable.SetPosition(instance["x"], instance["y"])
            variable.SetSize(instance["width"], instance["height"])
            self.AddBlock(variable)
            connectors = variable.GetConnectors()
            connectors["output"].SetPosition(wx.Point(*instance["connector"]["position"]))
            if instance["connector"]["negated"]:
                connectors["output"].SetNegated(True)
            if instance["connector"]["edge"]:
                connectors["output"].SetEdge(instance["connector"]["edge"])
        elif instance["type"] == "output":
            variable = FBD_Variable(self, OUTPUT, instance["name"], instance["value_type"], instance["id"], instance["executionOrder"])
            variable.SetPosition(instance["x"], instance["y"])
            variable.SetSize(instance["width"], instance["height"])
            self.AddBlock(variable)
            connectors = variable.GetConnectors()
            connectors["input"].SetPosition(wx.Point(*instance["connector"]["position"]))
            if instance["connector"]["negated"]:
                connectors["input"].SetNegated(True)
            if instance["connector"]["edge"]:
                connectors["input"].SetEdge(instance["connector"]["edge"])
            self.CreateWires(connectors["input"], instance["connector"]["links"], ids)
        elif instance["type"] == "inout":
            variable = FBD_Variable(self, INOUT, instance["name"], instance["value_type"], instance["id"], instance["executionOrder"])
            variable.SetPosition(instance["x"], instance["y"])
            variable.SetSize(instance["width"], instance["height"])
            self.AddBlock(variable)
            connectors = variable.GetConnectors()
            connectors["output"].SetPosition(wx.Point(*instance["connectors"]["output"]["position"]))
            connectors["input"].SetPosition(wx.Point(*instance["connectors"]["input"]["position"]))
            if instance["connectors"]["output"]["negated"]:
                connectors["output"].SetNegated(True)
            if instance["connectors"]["output"]["edge"]:
                connectors["output"].SetEdge(instance["connectors"]["output"]["edge"])
            if instance["connectors"]["input"]["negated"]:
                connectors["input"].SetNegated(True)
            if instance["connectors"]["input"]["edge"]:
                connectors["input"].SetEdge(instance["connectors"]["input"]["edge"])
            self.CreateWires(connectors["input"], instance["connectors"]["input"]["links"], ids)
        elif instance["type"] == "continuation":
            connection = FBD_Connector(self, CONTINUATION, instance["name"], instance["id"])
            connection.SetPosition(instance["x"], instance["y"])
            connection.SetSize(instance["width"], instance["height"])
            self.AddBlock(connection)
            connector = connection.GetConnector()
            connector.SetPosition(wx.Point(*instance["connector"]["position"]))
        elif instance["type"] == "connection":
            connection = FBD_Connector(self, CONNECTOR, instance["name"], instance["id"])
            connection.SetPosition(instance["x"], instance["y"])
            connection.SetSize(instance["width"], instance["height"])
            self.AddBlock(connection)
            connector = connection.GetConnector()
            connector.SetPosition(wx.Point(*instance["connector"]["position"]))
            self.CreateWires(connector, instance["connector"]["links"], ids)
        elif instance["type"] == "comment":
            comment = Comment(self, instance["content"], instance["id"])
            comment.SetPosition(instance["x"], instance["y"])
            comment.SetSize(instance["width"], instance["height"])
            self.AddComment(comment)
        elif instance["type"] == "leftPowerRail":
            leftpowerrail = LD_PowerRail(self, LEFTRAIL, instance["id"], [True for i in range(len(instance["connectors"]))])
            leftpowerrail.SetPosition(instance["x"], instance["y"])
            leftpowerrail.SetSize(instance["width"], instance["height"])
            self.AddBlock(leftpowerrail)
            connectors = leftpowerrail.GetConnectors()
            for i, connector in enumerate(instance["connectors"]):
                connectors[i].SetPosition(wx.Point(*connector["position"]))
        elif instance["type"] == "rightPowerRail":
            rightpowerrail = LD_PowerRail(self, RIGHTRAIL, instance["id"], [True for i in range(len(instance["connectors"]))])
            rightpowerrail.SetPosition(instance["x"], instance["y"])
            rightpowerrail.SetSize(instance["width"], instance["height"])
            self.AddBlock(rightpowerrail)
            connectors = rightpowerrail.GetConnectors()
            for i, connector in enumerate(instance["connectors"]):
                connectors[i].SetPosition(wx.Point(*connector["position"]))
                self.CreateWires(connectors[i], connector["links"], ids)
        elif instance["type"] == "contact":
            if instance["negated"]:
                negated = instance["negated"]
            else:
                negated = False
            if instance["edge"]:
                edge = instance["edge"]
            else:
                edge = "none"
            if negated and edge == "none":
                contact_type = CONTACT_REVERSE
            elif not negated and edge == "rising":
                contact_type = CONTACT_RISING
            elif not negated and edge == "falling":
                contact_type = CONTACT_FALLING
            else:
                contact_type = CONTACT_NORMAL
            contact = LD_Contact(self, contact_type, instance["name"], instance["id"])
            contact.SetPosition(instance["x"], instance["y"])
            contact.SetSize(instance["width"], instance["height"])
            self.AddBlock(contact)
            connectors = contact.GetConnectors()
            connectors["input"].SetPosition(wx.Point(*instance["connectors"]["input"]["position"]))
            self.CreateWires(connectors["input"], instance["connectors"]["input"]["links"], ids)
            connectors["output"].SetPosition(wx.Point(*instance["connectors"]["output"]["position"]))
        elif instance["type"] == "coil":
            if instance["negated"]:
                negated = instance["negated"]
            else:
                negated = False
            if instance["storage"]:
                storage = instance["storage"]
            else:
                storage = "none"
            if negated and storage == "none":
                coil_type = COIL_REVERSE
            elif not negated and storage == "set":
                coil_type = COIL_SET
            elif not negated and storage == "reset":
                coil_type = COIL_RESET
            else:
                coil_type = COIL_NORMAL
            coil = LD_Coil(self, coil_type, instance["name"], instance["id"])
            coil.SetPosition(instance["x"], instance["y"])
            coil.SetSize(instance["width"], instance["height"])
            self.AddBlock(coil)
            connectors = coil.GetConnectors()
            connectors["input"].SetPosition(wx.Point(*instance["connectors"]["input"]["position"]))
            self.CreateWires(connectors["input"], instance["connectors"]["input"]["links"], ids)
            connectors["output"].SetPosition(wx.Point(*instance["connectors"]["output"]["position"]))
        elif instance["type"] == "step":
            if instance["initial"]:
                initial = instance["initial"]
            else:
                initial = False
            step = SFC_Step(self, instance["name"], initial, instance["id"])
            step.SetPosition(instance["x"], instance["y"])
            step.SetSize(instance["width"], instance["height"])
            self.AddBlock(step)
            if "output" in instance["connectors"]:
                step.AddOutput()
            if "action" in instance["connectors"]:
                step.AddAction()
            connectors = step.GetConnectors()
            if connectors["input"]:
                connectors["input"].SetPosition(wx.Point(*instance["connectors"]["input"]["position"]))
                self.CreateWires(connectors["input"], instance["connectors"]["input"]["links"], ids)
            if connectors["output"]:
                connectors["output"].SetPosition(wx.Point(*instance["connectors"]["output"]["position"]))
            if connectors["action"]:
                connectors["action"].SetPosition(wx.Point(*instance["connectors"]["action"]["position"]))
        elif instance["type"] == "transition":
            transition = SFC_Transition(self, instance["condition_type"], instance["condition"], instance["priority"], instance["id"])
            transition.SetPosition(instance["x"], instance["y"])
            transition.SetSize(instance["width"], instance["height"])
            self.AddBlock(transition)
            connectors = transition.GetConnectors()
            connectors["input"].SetPosition(wx.Point(*instance["connectors"]["input"]["position"]))
            self.CreateWires(connectors["input"], instance["connectors"]["input"]["links"], ids)
            connectors["output"].SetPosition(wx.Point(*instance["connectors"]["output"]["position"]))
            if instance["condition_type"] == "connection":
                self.CreateWires(connectors["connection"], instance["connectors"]["connection"]["links"], ids)
        elif instance["type"] in ["selectionDivergence", "selectionConvergence", "simultaneousDivergence", "simultaneousConvergence"]:
            if instance["type"] == "selectionDivergence":
                divergence = SFC_Divergence(self, SELECTION_DIVERGENCE, 
                    len(instance["connectors"]["outputs"]), instance["id"])
            elif instance["type"] == "selectionConvergence":
                divergence = SFC_Divergence(self, SELECTION_CONVERGENCE, 
                    len(instance["connectors"]["inputs"]), instance["id"])
            elif instance["type"] == "simultaneousDivergence":
                divergence = SFC_Divergence(self, SIMULTANEOUS_DIVERGENCE, 
                    len(instance["connectors"]["outputs"]), instance["id"])
            else:
                divergence = SFC_Divergence(self, SIMULTANEOUS_CONVERGENCE, 
                    len(instance["connectors"]["inputs"]), instance["id"])
            divergence.SetPosition(instance["x"], instance["y"])
            divergence.SetSize(instance["width"], instance["height"])
            self.AddBlock(divergence)
            connectors = divergence.GetConnectors()
            for i, input_connector in enumerate(instance["connectors"]["inputs"]):
                connector = connectors["inputs"][i]
                connector.SetPosition(wx.Point(*input_connector["position"]))
                self.CreateWires(connector, input_connector["links"], ids)
            for i, output_connector in enumerate(instance["connectors"]["outputs"]):
                connector = connectors["outputs"][i]
                connector.SetPosition(wx.Point(*output_connector["position"]))
        elif instance["type"] == "jump":
            jump = SFC_Jump(self, instance["target"], instance["id"])
            jump.SetPosition(instance["x"], instance["y"])
            jump.SetSize(instance["width"], instance["height"])
            self.AddBlock(jump)
            connector = jump.GetConnector()
            connector.SetPosition(wx.Point(*instance["connector"]["position"]))
            self.CreateWires(connector, instance["connector"]["links"], ids)
        elif instance["type"] == "actionBlock":
            actionBlock = SFC_ActionBlock(self, instance["actions"], instance["id"])
            actionBlock.SetPosition(instance["x"], instance["y"])
            actionBlock.SetSize(instance["width"], instance["height"])
            self.AddBlock(actionBlock)
            connector = actionBlock.GetConnector()
            connector.SetPosition(wx.Point(*instance["connector"]["position"]))
            self.CreateWires(connector, instance["connector"]["links"], ids)
        else:
            connectors = {"inputs" : [], "outputs" : []}
            for input in instance["connectors"]["inputs"]:
                if input["negated"]:
                    connectors["inputs"].append((input["name"], None, "negated"))
                elif input["edge"]:
                    connectors["inputs"].append((input["name"], None, input["edge"]))
                else:
                    connectors["inputs"].append((input["name"], None, "none"))
            for output in instance["connectors"]["outputs"]:
                if output["negated"]:
                    connectors["outputs"].append((output["name"], None, "negated"))
                elif output["edge"]:
                    connectors["outputs"].append((output["name"], None, output["edge"]))
                else:
                    connectors["outputs"].append((output["name"], None, "none"))
            if instance["name"] != None:
                block = FBD_Block(self, instance["type"], instance["name"], instance["id"], len(instance["connectors"]["inputs"]), connectors=connectors, executionOrder=instance["executionOrder"])
            else:
                block = FBD_Block(self, instance["type"], "", instance["id"], len(instance["connectors"]["inputs"]), connectors=connectors, executionOrder=instance["executionOrder"])
            block.SetPosition(instance["x"], instance["y"])
            block.SetSize(instance["width"], instance["height"])
            self.AddBlock(block)
            connectors = block.GetConnectors()
            for i, input_connector in enumerate(instance["connectors"]["inputs"]):
                if i < len(connectors["inputs"]):
                    connector = connectors["inputs"][i]
                    connector.SetPosition(wx.Point(*input_connector["position"]))
                    if input_connector["negated"]:
                        connector.SetNegated(True)
                    if input_connector["edge"] != "none":
                        connector.SetEdge(input_connector["edge"])
                    self.CreateWires(connector, input_connector["links"], ids)
            for i, output_connector in enumerate(instance["connectors"]["outputs"]):
                if i < len(connectors["outputs"]):
                    connector = connectors["outputs"][i]
                    if output_connector["negated"]:
                        connector.SetNegated(True)
                    if output_connector["edge"] != "none":
                        connector.SetEdge(output_connector["edge"])
                    connector.SetPosition(wx.Point(*output_connector["position"]))
    
    def CreateWires(self, start_connector, links, ids):
        for link in links:
            refLocalId = link["refLocalId"]
            if refLocalId != None:
                if refLocalId not in ids:
                    new_instance = self.Controler.GetEditedElementInstanceInfos(self.TagName, refLocalId)
                    if new_instance:
                        self.loadInstance(new_instance, ids)
                connected = self.FindElementById(refLocalId)
                if connected:
                    points = link["points"]
                    end_connector = connected.GetConnector(wx.Point(points[-1][0], points[-1][1]), link["formalParameter"])
                    if end_connector:
                        wire = Wire(self)
                        wire.SetPoints(points)
                        start_connector.Connect((wire, 0), False)
                        end_connector.Connect((wire, -1), False)
                        wire.ConnectStartPoint(None, start_connector)
                        wire.ConnectEndPoint(None, end_connector)
                        self.AddWire(wire)

#-------------------------------------------------------------------------------
#                          Search Element functions
#-------------------------------------------------------------------------------

    def FindBlock(self, pos):
        for block in self.Blocks:
            if block.HitTest(pos) or block.TestHandle(pos) != (0, 0):
                return block
        return None
    
    def FindWire(self, pos):
        for wire in self.Wires:
            if wire.HitTest(pos) or wire.TestHandle(pos) != (0, 0):
                return wire
        return None
    
    def FindElement(self, pos, exclude_group = False):
        if self.SelectedElement and not (exclude_group and isinstance(self.SelectedElement, Graphic_Group)):
            if self.SelectedElement.HitTest(pos) or self.SelectedElement.TestHandle(pos) != (0, 0):
                return self.SelectedElement
        for element in self.GetElements():
            if element.HitTest(pos) or element.TestHandle(pos) != (0, 0):
                return element
        return None
    
    def FindBlockConnector(self, pos, exclude = True):
        for block in self.Blocks:
            result = block.TestConnector(pos, exclude)
            if result:
                return result
        return None
    
    def FindElementById(self, id):
        for element in self.Blocks:
            if element.GetId() == id:
                return element
        for element in self.Comments:
            if element.GetId() == id:
                return element
        return None
    
    def SearchElements(self, bbox):
        elements = []
        for element in self.GetElements():
            if element.IsInSelection(bbox):
                elements.append(element)
        return elements

#-------------------------------------------------------------------------------
#                           Popup menu functions
#-------------------------------------------------------------------------------

    def PopupBlockMenu(self, connector = None):
        if connector is not None and connector.IsCompatible("BOOL"):
            type = self.Controler.GetEditedElementType(self.TagName)
            self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS0, True)
            self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS1, True)
            self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS2, type != "function")
            self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS3, type != "function")
        else:
            self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS0, False)
            self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS1, False)
            self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS2, False)
            self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS3, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS5, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS6, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS8, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS9, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS14, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS16, self.SelectedElement.GetType() in self.Controler.GetProjectPouNames())
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS17, True)
        if connector:
            if connector.IsNegated():
                self.ContextualMenu.Check(ID_VIEWERCONTEXTUALMENUITEMS1, True)
            elif connector.GetEdge() == "rising":
                self.ContextualMenu.Check(ID_VIEWERCONTEXTUALMENUITEMS2, True)
            elif connector.GetEdge() == "falling":
                self.ContextualMenu.Check(ID_VIEWERCONTEXTUALMENUITEMS3, True)
            else:
                self.ContextualMenu.Check(ID_VIEWERCONTEXTUALMENUITEMS0, True)
        self.PopupMenu(self.ContextualMenu)
    
    def PopupWireMenu(self):
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS0, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS1, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS2, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS3, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS5, True)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS6, True)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS8, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS9, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS14, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS16, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS17, True)
        self.PopupMenu(self.ContextualMenu)
    
    def PopupDivergenceMenu(self, connector):
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS0, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS1, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS2, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS3, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS5, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS6, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS8, True)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS9, connector)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS14, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS16, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS17, True)
        self.PopupMenu(self.ContextualMenu)
    
    def PopupGroupMenu(self):
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS0, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS1, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS2, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS3, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS5, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS6, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS8, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS9, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS14, True)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS16, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS17, True)
        self.PopupMenu(self.ContextualMenu)
    
    def PopupDefaultMenu(self, block = True):
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS0, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS1, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS2, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS3, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS5, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS6, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS8, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS9, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS14, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS16, False)
        self.ContextualMenu.Enable(ID_VIEWERCONTEXTUALMENUITEMS17, block)
        self.PopupMenu(self.ContextualMenu)

#-------------------------------------------------------------------------------
#                            Menu items functions
#-------------------------------------------------------------------------------

    def OnAlignLeftMenu(self, event):
        if self.SelectedElement and isinstance(self.SelectedElement, Graphic_Group):
            self.SelectedElement.AlignElements(ALIGN_LEFT, None)
            self.RefreshBuffer()
            self.Refresh(False)
        event.Skip()
    
    def OnAlignCenterMenu(self, event):
        if self.SelectedElement and isinstance(self.SelectedElement, Graphic_Group):
            self.SelectedElement.AlignElements(ALIGN_CENTER, None)
            self.RefreshBuffer()
            self.Refresh(False)
        event.Skip()
    
    def OnAlignRightMenu(self, event):
        if self.SelectedElement and isinstance(self.SelectedElement, Graphic_Group):
            self.SelectedElement.AlignElements(ALIGN_RIGHT, None)
            self.RefreshBuffer()
            self.Refresh(False)
        event.Skip()
    
    def OnAlignTopMenu(self, event):
        if self.SelectedElement and isinstance(self.SelectedElement, Graphic_Group):
            self.SelectedElement.AlignElements(None, ALIGN_TOP)
            self.RefreshBuffer()
            self.Refresh(False)
        event.Skip()
    
    def OnAlignMiddleMenu(self, event):
        if self.SelectedElement and isinstance(self.SelectedElement, Graphic_Group):
            self.SelectedElement.AlignElements(None, ALIGN_MIDDLE)
            self.RefreshBuffer()
            self.Refresh(False)
        event.Skip()
    
    def OnAlignBottomMenu(self, event):
        if self.SelectedElement and isinstance(self.SelectedElement, Graphic_Group):
            self.SelectedElement.AlignElements(None, ALIGN_BOTTOM)
            self.RefreshBuffer()
            self.Refresh(False)
        event.Skip()
        
    def OnNoModifierMenu(self, event):
        if self.SelectedElement and self.IsBlock(self.SelectedElement):
            self.SelectedElement.SetConnectorNegated(False)
            self.RefreshBuffer()
        event.Skip()
    
    def OnNegatedMenu(self, event):
        if self.SelectedElement and self.IsBlock(self.SelectedElement):
            self.SelectedElement.SetConnectorNegated(True)
            self.RefreshBuffer()
        event.Skip()

    def OnRisingEdgeMenu(self, event):
        if self.SelectedElement and self.IsBlock(self.SelectedElement):
            self.SelectedElement.SetConnectorEdge("rising")
            self.RefreshBuffer()
        event.Skip()

    def OnFallingEdgeMenu(self, event):
        if self.SelectedElement and self.IsBlock(self.SelectedElement):
            self.SelectedElement.SetConnectorEdge("falling")
            self.RefreshBuffer()
        event.Skip()

    def OnAddSegmentMenu(self, event):
        if self.SelectedElement and self.IsWire(self.SelectedElement):
            self.SelectedElement.AddSegment()
        event.Skip()

    def OnDeleteSegmentMenu(self, event):
        if self.SelectedElement and self.IsWire(self.SelectedElement):
            self.SelectedElement.DeleteSegment()
        event.Skip()

    def OnAddBranchMenu(self, event):
        if self.SelectedElement and self.IsBlock(self.SelectedElement):
            self.AddDivergenceBranch(self.SelectedElement)
            self.RefreshBuffer()
        event.Skip()

    def OnDeleteBranchMenu(self, event):
        if self.SelectedElement and self.IsBlock(self.SelectedElement):
            self.RemoveDivergenceBranch(self.SelectedElement)
            self.RefreshBuffer()
        event.Skip()

    def OnEditBlockMenu(self, event):
        if self.SelectedElement:
            self.ParentWindow.EditProjectElement(ITEM_POU, "P::%s"%self.SelectedElement.GetType())
        event.Skip()

    def OnDeleteMenu(self, event):
        if self.SelectedElement:
            self.SelectedElement.Delete()
            self.SelectedElement = None
            self.RefreshBuffer()
        event.Skip()

    def OnClearExecutionOrderMenu(self, event):
        self.Controler.ClearEditedElementExecutionOrder(self.TagName)
        self.RefreshBuffer()
        self.RefreshView()
        
    def OnResetExecutionOrderMenu(self, event):
        self.Controler.ResetEditedElementExecutionOrder(self.TagName)
        self.RefreshBuffer()
        self.RefreshView()

#-------------------------------------------------------------------------------
#                          Mouse event functions
#-------------------------------------------------------------------------------

    def OnViewerLeftDown(self, event):
        if self.Mode == MODE_SELECTION:
            dc = self.GetLogicalDC()
            pos = event.GetLogicalPosition(dc)
            if event.ControlDown() and self.SelectedElement:
                element = self.FindElement(pos, True)
                if element:
                    if isinstance(self.SelectedElement, Graphic_Group):
                        self.SelectedElement.SetSelected(False)
                        self.SelectedElement.SelectElement(element)
                    elif self.SelectedElement:
                        group = Graphic_Group(self)
                        group.SelectElement(self.SelectedElement)
                        group.SelectElement(element)
                        self.SelectedElement = group
                    elements = self.SelectedElement.GetElements()
                    if len(elements) == 0:
                        self.SelectedElement = element
                    elif len(elements) == 1:
                        self.SelectedElement = elements[0]
                    self.SelectedElement.SetSelected(True)
            else:
                element = self.FindElement(pos)
                if element is None or element.TestHandle(pos) == (0, 0):
                    connector = self.FindBlockConnector(pos)
                else:
                    connector = None
                if self.DrawingWire:
                    self.DrawingWire = False
                    if self.SelectedElement is not None:
                        if connector:
                            event.Dragging = lambda : True
                            self.SelectedElement.OnMotion(event, self.GetLogicalDC(), self.Scaling)
                        if self.SelectedElement.EndConnected is not None:
                            self.SelectedElement.ResetPoints()
                            self.SelectedElement.GeneratePoints()
                            self.SelectedElement.RefreshModel()
                            self.SelectedElement.SetSelected(True)
                            element = self.SelectedElement
                            self.RefreshBuffer()
                        else:
                            rect = self.SelectedElement.GetRedrawRect()
                            self.SelectedElement.Delete()
                            self.SelectedElement = None
                            element = None
                            self.RefreshRect(self.GetScrolledRect(rect), False)
                elif connector:
                    self.DrawingWire = True
                    pos = GetScaledEventPosition(event, self.GetLogicalDC(), self.Scaling)
                    if (connector.GetDirection() == EAST):
                        wire = Wire(self, [wx.Point(pos.x, pos.y), EAST], [wx.Point(pos.x, pos.y), WEST])
                    else:
                        wire = Wire(self, [wx.Point(pos.x, pos.y), WEST], [wx.Point(pos.x, pos.y), EAST])
                    wire.oldPos = pos
                    wire.Handle = (HANDLE_POINT, 0)
                    wire.ProcessDragging(0, 0, False, self.Scaling)
                    wire.Handle = (HANDLE_POINT, 1)
                    self.AddWire(wire)
                    if self.SelectedElement:
                        self.SelectedElement.SetSelected(False)
                    self.SelectedElement = wire
                    self.SelectedElement.Refresh()
                else:
                    if self.SelectedElement and self.SelectedElement != element:
                        self.SelectedElement.SetSelected(False)
                        self.SelectedElement = None
                    if element:
                        self.SelectedElement = element
                        self.SelectedElement.OnLeftDown(event, dc, self.Scaling)
                        self.SelectedElement.Refresh()
                    else:
                        self.rubberBand.Reset()
                        self.rubberBand.OnLeftDown(event, dc, self.Scaling)
        elif self.Mode in [MODE_BLOCK, MODE_VARIABLE, MODE_CONNECTION, MODE_COMMENT, 
                           MODE_CONTACT, MODE_COIL, MODE_POWERRAIL, MODE_INITIALSTEP, 
                           MODE_STEP, MODE_TRANSITION, MODE_DIVERGENCE, MODE_JUMP, MODE_ACTION]:
            self.rubberBand.Reset()
            self.rubberBand.OnLeftDown(event, self.GetLogicalDC(), self.Scaling)
        event.Skip()

    def OnViewerLeftUp(self, event):
        if self.rubberBand.IsShown():
            if self.Mode == MODE_SELECTION:
                elements = self.SearchElements(self.rubberBand.GetCurrentExtent())
                self.rubberBand.OnLeftUp(event, self.GetLogicalDC(), self.Scaling)
                if len(elements) == 1:
                    self.SelectedElement = elements[0]
                    self.SelectedElement.SetSelected(True)
                elif len(elements) > 1:
                    self.SelectedElement = Graphic_Group(self)
                    self.SelectedElement.SetElements(elements)
                    self.SelectedElement.SetSelected(True)
            else:
                bbox = self.rubberBand.GetCurrentExtent()
                self.rubberBand.OnLeftUp(event, self.GetLogicalDC(), self.Scaling)                
                if self.Mode == MODE_BLOCK:
                    wx.CallAfter(self.AddNewBlock, bbox)
                elif self.Mode == MODE_VARIABLE:
                    wx.CallAfter(self.AddNewVariable, bbox)
                elif self.Mode == MODE_CONNECTION:
                    wx.CallAfter(self.AddNewConnection, bbox)
                elif self.Mode == MODE_COMMENT:
                    wx.CallAfter(self.AddNewComment, bbox)
                elif self.Mode == MODE_CONTACT:
                    wx.CallAfter(self.AddNewContact, bbox)
                elif self.Mode == MODE_COIL:
                    wx.CallAfter(self.AddNewCoil, bbox)
                elif self.Mode == MODE_POWERRAIL:
                    wx.CallAfter(self.AddNewPowerRail, bbox)
                elif self.Mode == MODE_INITIALSTEP:
                    wx.CallAfter(self.AddNewStep, bbox, True)
                elif self.Mode == MODE_STEP:
                    wx.CallAfter(self.AddNewStep, bbox, False)
                elif self.Mode == MODE_TRANSITION:
                    wx.CallAfter(self.AddNewTransition, bbox)
                elif self.Mode == MODE_DIVERGENCE:
                    wx.CallAfter(self.AddNewDivergence, bbox)
                elif self.Mode == MODE_JUMP:
                    wx.CallAfter(self.AddNewJump, bbox)
                elif self.Mode == MODE_ACTION:
                    wx.CallAfter(self.AddNewActionBlock, bbox)
        elif self.Mode == MODE_SELECTION and self.SelectedElement:
            dc = self.GetLogicalDC()
            if self.DrawingWire:
                pos = event.GetLogicalPosition(dc)
                connector = self.FindBlockConnector(pos)
                if self.SelectedElement.EndConnected is not None:
                    self.DrawingWire = False
                    self.SelectedElement.StartConnected.HighlightParentBlock(False)
                    self.SelectedElement.EndConnected.HighlightParentBlock(False)
                    self.SelectedElement.ResetPoints()
                    self.SelectedElement.OnMotion(event, dc, self.Scaling)
                    self.SelectedElement.GeneratePoints()
                    self.SelectedElement.RefreshModel()
                    if self.HighlightedElement is not None:
                        self.HighlightedElement.SetHighlighted(False)
                        self.HighlightedElement = None
                    self.SelectedElement.SetHighlighted(True)
                    self.HighlightedElement = self.SelectedElement
                    self.SelectedElement.SetSelected(True)
                    self.RefreshBuffer()
                elif connector is None or self.SelectedElement.GetDragging():
                    self.DrawingWire = False
                    rect = self.SelectedElement.GetRedrawRect()
                    self.SelectedElement.Delete()
                    self.SelectedElement = None
                    self.RefreshRect(self.GetScrolledRect(rect), False)
            else:
                self.SelectedElement.OnLeftUp(event, dc, self.Scaling)
                wx.CallAfter(self.SetCursor, wx.NullCursor)
                self.SelectedElement.Refresh()
        if self.Mode != MODE_SELECTION and not self.SavedMode:
            wx.CallAfter(self.ParentWindow.ResetCurrentMode)
        event.Skip()
    
    def OnViewerRightDown(self, event):
        if self.Mode == MODE_SELECTION:
            dc = self.GetLogicalDC()
            pos = event.GetLogicalPosition(dc)
            element = self.FindElement(pos)
            if self.SelectedElement and self.SelectedElement != element:
                self.SelectedElement.SetSelected(False)
                self.SelectedElement = None
            if element:
                self.SelectedElement = element
                self.SelectedElement.OnRightDown(event, dc, self.Scaling)
                self.SelectedElement.Refresh()
        event.Skip()
    
    def OnViewerRightUp(self, event):
        dc = self.GetLogicalDC()
        if self.SelectedElement:
            self.SelectedElement.OnRightUp(event, dc, self.Scaling)
            wx.CallAfter(self.SetCursor, wx.NullCursor)
            self.SelectedElement.Refresh()
        else:
            self.PopupDefaultMenu(False)
        event.Skip()
    
    def OnViewerLeftDClick(self, event):
        if self.Mode == MODE_SELECTION and self.SelectedElement:
            if event.ControlDown() and self.IsBlock(self.SelectedElement) and self.SelectedElement.GetType() in self.Controler.GetProjectPouNames():
                self.ParentWindow.EditProjectElement(ITEM_POU, self.SelectedElement.GetType())
            else:
                self.SelectedElement.OnLeftDClick(event, self.GetLogicalDC(), self.Scaling)
        event.Skip()
    
    def OnViewerMotion(self, event):
        refresh = False
        dc = self.GetLogicalDC()
        pos = GetScaledEventPosition(event, dc, self.Scaling)
        if not event.Dragging():
            highlighted = self.FindElement(pos) 
            if self.HighlightedElement is not None and self.HighlightedElement != highlighted:
                self.HighlightedElement.SetHighlighted(False)
                self.HighlightedElement = None
            if highlighted is not None and self.HighlightedElement != highlighted:
                highlighted.SetHighlighted(True)
            self.HighlightedElement = highlighted
        if self.rubberBand.IsShown():
            self.rubberBand.OnMotion(event, dc, self.Scaling)
        elif self.Mode == MODE_SELECTION and self.SelectedElement:
            if self.DrawingWire:
                connector = self.FindBlockConnector(pos, False)
                if not connector or self.SelectedElement.EndConnected == None:
                    self.SelectedElement.ResetPoints()
                    movex, movey = self.SelectedElement.OnMotion(event, dc, self.Scaling)
                    self.SelectedElement.GeneratePoints()
                    if movex != 0 or movey != 0:
                        self.RefreshRect(self.GetScrolledRect(self.SelectedElement.GetRedrawRect(movex, movey)), False)
            else:
                movex, movey = self.SelectedElement.OnMotion(event, dc, self.Scaling)
                if movex != 0 or movey != 0:
                    self.RefreshRect(self.GetScrolledRect(self.SelectedElement.GetRedrawRect(movex, movey)), False)
        self.UpdateScrollPos(event)
        event.Skip()

    def OnLeaveViewer(self, event):
        if self.SelectedElement and self.SelectedElement.GetDragging():
            event.Skip()
        elif self.HighlightedElement is not None:
            self.HighlightedElement.SetHighlighted(False)
            self.HighlightedElement = None
        event.Skip()

    def UpdateScrollPos(self, event):
        if (event.Dragging() and self.SelectedElement) or self.rubberBand.IsShown():
            position = event.GetPosition()
            move_window = wx.Point()
            window_size = self.GetClientSize()
            xstart, ystart = self.GetViewStart()
            if position.x < SCROLL_ZONE and xstart > 0:
                move_window.x = -1
            elif position.x > window_size[0] - SCROLL_ZONE:
                move_window.x = 1
            if position.y < SCROLL_ZONE and ystart > 0:
                move_window.y = -1
            elif position.y > window_size[1] - SCROLL_ZONE:
                move_window.y = 1
            if move_window.x != 0 or move_window.y != 0:
                self.Scroll(xstart + move_window.x, ystart + move_window.y)
            self.RefreshScrollBars()

#-------------------------------------------------------------------------------
#                          Keyboard event functions
#-------------------------------------------------------------------------------

    def OnChar(self, event):
        xpos, ypos = self.GetScrollPos(wx.HORIZONTAL), self.GetScrollPos(wx.VERTICAL)
        xmax = self.GetScrollRange(wx.HORIZONTAL) - self.GetScrollThumb(wx.HORIZONTAL)
        ymax = self.GetScrollRange(wx.VERTICAL) - self.GetScrollThumb(wx.VERTICAL)
        keycode = event.GetKeyCode()
        if self.Scaling:
            scaling = self.Scaling
        else:
            scaling = (8, 8)
        if keycode == wx.WXK_DELETE and self.SelectedElement:
            rect = self.SelectedElement.GetRedrawRect(1, 1)
            self.SelectedElement.Delete()
            self.SelectedElement = None
            self.RefreshBuffer()
            self.RefreshScrollBars()
            self.SetCursor(wx.NullCursor)
            self.RefreshRect(self.GetScrolledRect(rect), False)
        elif keycode == wx.WXK_RETURN and self.SelectedElement:
            self.SelectedElement.OnLeftDClick(event, self.GetLogicalDC(), self.Scaling)
        elif keycode == wx.WXK_LEFT:
            if event.ControlDown() and event.ShiftDown():
                self.Scroll(0, ypos)
            elif event.ControlDown():
                event.Skip()
            elif self.SelectedElement:
                self.SelectedElement.Move(-scaling[0], 0)
                self.SelectedElement.RefreshModel()
                self.RefreshBuffer()
                self.RefreshScrollBars()
                self.RefreshRect(self.GetScrolledRect(self.SelectedElement.GetRedrawRect(-scaling[0], 0)), False)
        elif keycode == wx.WXK_RIGHT:
            if event.ControlDown() and event.ShiftDown():
                self.Scroll(xmax, ypos)
            elif event.ControlDown():
                event.Skip()
            elif self.SelectedElement:
                self.SelectedElement.Move(scaling[0], 0)
                self.SelectedElement.RefreshModel()
                self.RefreshBuffer()
                self.RefreshScrollBars()
                self.RefreshRect(self.GetScrolledRect(self.SelectedElement.GetRedrawRect(scaling[0], 0)), False)
        elif keycode == wx.WXK_UP:
            if event.ControlDown() and event.ShiftDown():
                self.Scroll(xpos, 0)
            elif event.ControlDown():
                event.Skip()
            elif self.SelectedElement:
                self.SelectedElement.Move(0, -scaling[1])
                self.SelectedElement.RefreshModel()
                self.RefreshBuffer()
                self.RefreshScrollBars()
                self.RefreshRect(self.GetScrolledRect(self.SelectedElement.GetRedrawRect(0, -scaling[1])), False)
        elif keycode == wx.WXK_DOWN:
            if event.ControlDown() and event.ShiftDown():
                self.Scroll(xpos, ymax)
            elif event.ControlDown():
                event.Skip()
            elif self.SelectedElement:
                self.SelectedElement.Move(0, scaling[1])
                self.SelectedElement.RefreshModel()
                self.RefreshBuffer()
                self.RefreshScrollBars()
                self.RefreshRect(self.GetScrolledRect(self.SelectedElement.GetRedrawRect(0, scaling[1])), False)
        elif keycode == wx.WXK_SPACE and self.SelectedElement is not None and self.SelectedElement.Dragging:
            self.CopyBlock(self.SelectedElement, wx.Point(*self.SelectedElement.GetPosition()))
            self.RefreshBuffer()
            self.RefreshScrollBars()
            self.ParentWindow.RefreshVariablePanel(self.TagName)
            self.SelectedElement.Refresh()
        else:
            event.Skip()

#-------------------------------------------------------------------------------
#                          Model adding functions
#-------------------------------------------------------------------------------

    def GetScaledSize(self, width, height):
        if self.Scaling is not None:
            width = round(float(width) / float(self.Scaling[0]) + 0.4) * self.Scaling[0]
            height = round(float(height) / float(self.Scaling[1]) + 0.4) * self.Scaling[1]
        return width, height

    def AddNewBlock(self, bbox):
        dialog = BlockPropertiesDialog(self.ParentWindow)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetBlockList(self.Controler.GetBlockTypes(self.TagName))
        dialog.SetPouNames(self.Controler.GetProjectPouNames())
        dialog.SetPouElementNames(self.Controler.GetEditedElementVariables(self.TagName))
        dialog.SetMinBlockSize((bbox.width, bbox.height))
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            if "name" in values:
                block = FBD_Block(self, values["type"], values["name"], id, values["extension"], values["inputs"])
            else:
                block = FBD_Block(self, values["type"], "", id, values["extension"], values["inputs"])
            block.SetPosition(bbox.x, bbox.y)
            block.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            self.AddBlock(block)
            self.Controler.AddEditedElementBlock(self.TagName, id, values["type"], values.get("name", None))
            self.RefreshBlockModel(block)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            self.ParentWindow.RefreshVariablePanel(self.TagName)
            block.Refresh()
        dialog.Destroy()
    
    def AddNewVariable(self, bbox):
        dialog = VariablePropertiesDialog(self.ParentWindow)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetMinVariableSize((bbox.width, bbox.height))
        varlist = []
        vars = self.Controler.GetEditedElementInterfaceVars(self.TagName)
        if vars:
            for var in vars:
                if var["Edit"]:
                    varlist.append((var["Name"], var["Class"], var["Type"]))
        returntype = self.Controler.GetEditedElementInterfaceReturnType(self.TagName)
        if returntype:
            varlist.append((self.Controler.GetEditedElementName(self.TagName), "Output", returntype))
        dialog.SetVariables(varlist)
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            variable = FBD_Variable(self, values["type"], values["name"], values["value_type"], id)
            variable.SetPosition(bbox.x, bbox.y)
            variable.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            self.AddBlock(variable)
            self.Controler.AddEditedElementVariable(self.TagName, id, values["type"])
            self.RefreshVariableModel(variable)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            variable.Refresh()
        dialog.Destroy()

    def AddNewConnection(self, bbox):
        dialog = ConnectionPropertiesDialog(self.ParentWindow)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetMinConnectionSize((bbox.width, bbox.height))
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            connection = FBD_Connector(self, values["type"], values["name"], id)
            connection.SetPosition(bbox.x, bbox.y)
            connection.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            self.AddBlock(connection)
            self.Controler.AddEditedElementConnection(self.TagName, id, values["type"])
            self.RefreshConnectionModel(connection)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            connection.Refresh()
        dialog.Destroy()

    def AddNewComment(self, bbox):
        if wx.VERSION >= (2, 5, 0):
            dialog = wx.TextEntryDialog(self.ParentWindow, "Edit comment", "Please enter comment text", "", wx.OK|wx.CANCEL|wx.TE_MULTILINE)
        else:
            dialog = wx.TextEntryDialog(self.ParentWindow, "Edit comment", "Please enter comment text", "", wx.OK|wx.CANCEL)
        if dialog.ShowModal() == wx.ID_OK:
            value = dialog.GetValue()
            id = self.GetNewId()
            comment = Comment(self, value, id)
            comment.SetPosition(bbox.x, bbox.y)
            min_width, min_height = comment.GetMinSize()
            comment.SetSize(*self.GetScaledSize(max(min_width,bbox.width),max(min_height,bbox.height)))
            self.AddComment(comment)
            self.Controler.AddEditedElementComment(self.TagName, id)
            self.RefreshCommentModel(comment)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            comment.Refresh()
        dialog.Destroy()

    def AddNewContact(self, bbox):
        dialog = LDElementDialog(self.ParentWindow, "contact")
        dialog.SetPreviewFont(self.GetFont())
        varlist = []
        vars = self.Controler.GetEditedElementInterfaceVars(self.TagName)
        if vars:
            for var in vars:
                if var["Class"] != "Output" and var["Type"] == "BOOL":
                    varlist.append(var["Name"])
        dialog.SetVariables(varlist)
        dialog.SetValues({"name":"","type":CONTACT_NORMAL})
        dialog.SetElementSize((bbox.width, bbox.height))
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            contact = LD_Contact(self, values["type"], values["name"], id)
            contact.SetPosition(bbox.x, bbox.y)
            contact.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            self.AddBlock(contact)
            self.Controler.AddEditedElementContact(self.TagName, id)
            self.RefreshContactModel(contact)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            contact.Refresh()
        dialog.Destroy()

    def AddNewCoil(self, bbox):
        dialog = LDElementDialog(self.ParentWindow, "coil")
        dialog.SetPreviewFont(self.GetFont())
        varlist = []
        vars = self.Controler.GetEditedElementInterfaceVars(self.TagName)
        if vars:
            for var in vars:
                if var["Class"] != "Input" and var["Type"] == "BOOL":
                    varlist.append(var["Name"])
        returntype = self.Controler.GetEditedElementInterfaceReturnType(self.TagName)
        if returntype == "BOOL":
            varlist.append(self.Controler.GetEditedElementName(self.TagName))
        dialog.SetVariables(varlist)
        dialog.SetValues({"name":"","type":COIL_NORMAL})
        dialog.SetElementSize((bbox.width, bbox.height))
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            coil = LD_Coil(self, values["type"], values["name"], id)
            coil.SetPosition(bbox.x, bbox.y)
            coil.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            self.AddBlock(coil)
            self.Controler.AddEditedElementCoil(self.TagName, id)
            self.RefreshCoilModel(coil)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            coil.Refresh()
        dialog.Destroy()

    def AddNewPowerRail(self, bbox):
        dialog = LDPowerRailDialog(self.ParentWindow)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetMinSize((bbox.width, bbox.height))
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            powerrail = LD_PowerRail(self, values["type"], id, [True for i in xrange(values["number"])])
            powerrail.SetPosition(bbox.x, bbox.y)
            powerrail.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            self.AddBlock(powerrail)
            self.Controler.AddEditedElementPowerRail(self.TagName, id, values["type"])
            self.RefreshPowerRailModel(powerrail)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            powerrail.Refresh()
        dialog.Destroy()

    def AddNewStep(self, bbox, initial = False):
        dialog = StepContentDialog(self.ParentWindow, initial)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetPouNames(self.Controler.GetProjectPouNames())
        dialog.SetVariables(self.Controler.GetEditedElementInterfaceVars(self.TagName))
        dialog.SetStepNames([block.GetName() for block in self.Blocks if isinstance(block, SFC_Step)])
        dialog.SetMinStepSize((bbox.width, bbox.height))
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            step = SFC_Step(self, values["name"], initial, id)
            if values["input"]:
                step.AddInput()
            else:
                step.RemoveInput()
            if values["output"]:
                step.AddOutput()
            else:
                step.RemoveOutput()
            if values["action"]:
                step.AddAction()    
            else:
                step.RemoveAction()
            step.SetPosition(bbox.x, bbox.y)
            min_width, min_height = step.GetMinSize()
            step.SetSize(*self.GetScaledSize(max(bbox.width, min_width), max(bbox.height, min_height)))
            self.AddBlock(step)
            self.Controler.AddEditedElementStep(self.TagName, id)
            self.RefreshStepModel(step)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            step.Refresh()
        dialog.Destroy()

    def AddNewTransition(self, bbox):
        dialog = TransitionContentDialog(self.ParentWindow, self.GetDrawingMode() == FREEDRAWING_MODE)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetTransitions(self.Controler.GetEditedElementTransitions(self.TagName))
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            transition = SFC_Transition(self, values["type"], values["value"], values["priority"], id)
            transition.SetPosition(bbox.x, bbox.y)
            min_width, min_height = transition.GetMinSize()
            transition.SetSize(*self.GetScaledSize(max(bbox.width, min_width), max(bbox.height, min_height)))
            self.AddBlock(transition)
            self.Controler.AddEditedElementTransition(self.TagName, id)
            self.RefreshTransitionModel(transition)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            transition.Refresh()
        dialog.Destroy()

    def AddNewDivergence(self, bbox):
        dialog = DivergenceCreateDialog(self.ParentWindow)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetMinSize((bbox.width, bbox.height))
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            values = dialog.GetValues()
            divergence = SFC_Divergence(self, values["type"], values["number"], id)
            divergence.SetPosition(bbox.x, bbox.y)
            min_width, min_height = divergence.GetMinSize(True)
            divergence.SetSize(*self.GetScaledSize(max(bbox.width, min_width), max(bbox.height, min_height)))
            self.AddBlock(divergence)
            self.Controler.AddEditedElementDivergence(self.TagName, id, values["type"])
            self.RefreshDivergenceModel(divergence)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            divergence.Refresh()
        dialog.Destroy()

    def AddNewJump(self, bbox):
        choices = []
        for block in self.Blocks:
            if isinstance(block, SFC_Step):
                choices.append(block.GetName())
        dialog = wx.SingleChoiceDialog(self.ParentWindow, "Add a new jump", "Please choose a target", choices, wx.OK|wx.CANCEL)
        if dialog.ShowModal() == wx.ID_OK:
            id = self.GetNewId()
            value = dialog.GetStringSelection()
            jump = SFC_Jump(self, value, id)
            jump.SetPosition(bbox.x, bbox.y)
            min_width, min_height = jump.GetMinSize()
            jump.SetSize(*self.GetScaledSize(max(bbox.width, min_width), max(bbox.height, min_height)))
            self.AddBlock(jump)
            self.Controler.AddEditedElementJump(self.TagName, id)
            self.RefreshJumpModel(jump)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            jump.Refresh()
        dialog.Destroy()

    def AddNewActionBlock(self, bbox):
        dialog = ActionBlockDialog(self.ParentWindow)
        dialog.SetQualifierList(self.Controler.GetQualifierTypes())
        dialog.SetActionList(self.Controler.GetEditedElementActions(self.TagName))
        dialog.SetVariableList(self.Controler.GetEditedElementInterfaceVars(self.TagName))
        if dialog.ShowModal() == wx.ID_OK:
            actions = dialog.GetValues()
            id = self.GetNewId()
            actionblock = SFC_ActionBlock(self, actions, id)
            actionblock.SetPosition(bbox.x, bbox.y)
            min_width, min_height = actionblock.GetMinSize()
            actionblock.SetSize(*self.GetScaledSize(max(bbox.width, min_width), max(bbox.height, min_height)))
            self.AddBlock(actionblock)
            self.Controler.AddEditedElementActionBlock(self.TagName, id)
            self.RefreshActionBlockModel(actionblock)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            actionblock.Refresh()
        dialog.Destroy()

#-------------------------------------------------------------------------------
#                          Edit element content functions
#-------------------------------------------------------------------------------

    def EditBlockContent(self, block):
        dialog = BlockPropertiesDialog(self.ParentWindow)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetBlockList(self.Controler.GetBlockTypes(self.TagName))
        dialog.SetPouNames(self.Controler.GetProjectPouNames())
        variable_names = self.Controler.GetEditedElementVariables(self.TagName)
        if block.GetName() != "":
            variable_names.remove(block.GetName())
        dialog.SetPouElementNames(variable_names)
        dialog.SetMinBlockSize(block.GetSize())
        old_values = {"name" : block.GetName(), "type" : block.GetType(), "inputs" : block.GetInputTypes(), 
            "executionOrder" : block.GetExecutionOrder(), "extension" : block.GetExtension()}
        dialog.SetValues(old_values)
        if dialog.ShowModal() == wx.ID_OK:
            new_values = dialog.GetValues()
            rect = block.GetRedrawRect(1, 1)
            if "name" in new_values:
                block.SetName(new_values["name"])
            else:
                block.SetName("")
            block.SetSize(*self.GetScaledSize(new_values["width"], new_values["height"]))
            block.SetType(new_values["type"], new_values["extension"])
            block.SetExecutionOrder(new_values["executionOrder"])
            rect = rect.Union(block.GetRedrawRect())
            self.RefreshBlockModel(block)
            if old_values["executionOrder"] != new_values["executionOrder"]:
                self.RefreshView()
            self.RefreshBuffer()
            self.RefreshScrollBars()
            self.ParentWindow.RefreshVariablePanel(self.TagName)
            block.Refresh(rect)
        dialog.Destroy()

    def EditVariableContent(self, variable):
        dialog = VariablePropertiesDialog(self.ParentWindow)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetMinVariableSize(variable.GetSize())
        varlist = []
        vars = self.Controler.GetEditedElementInterfaceVars(self.TagName)
        if vars:
            for var in vars:
                if var["Edit"]:
                    varlist.append((var["Name"], var["Class"], var["Type"]))
        returntype = self.Controler.GetEditedElementInterfaceReturnType(self.TagName)
        if returntype:
            varlist.append((self.Controler.GetEditedElementName(self.TagName), "Output", returntype))
        dialog.SetVariables(varlist)
        old_values = {"name" : variable.GetName(), "type" : variable.GetType(), 
            "executionOrder" : variable.GetExecutionOrder()}
        dialog.SetValues(old_values)
        if dialog.ShowModal() == wx.ID_OK:
            new_values = dialog.GetValues()
            rect = variable.GetRedrawRect(1, 1)
            variable.SetName(new_values["name"])
            variable.SetType(new_values["type"], new_values["value_type"])
            variable.SetSize(*self.GetScaledSize(new_values["width"], new_values["height"]))
            variable.SetExecutionOrder(new_values["executionOrder"])
            rect = rect.Union(variable.GetRedrawRect())
            if old_values["type"] != new_values["type"]:
                id = variable.GetId()
                self.Controler.RemoveEditedElementInstance(self.TagName, id)
                self.Controler.AddEditedElementVariable(self.TagName, id, new_values["type"])
            self.RefreshVariableModel(variable)
            if old_values["executionOrder"] != new_values["executionOrder"]:
                self.RefreshView()
            self.RefreshBuffer()
            self.RefreshScrollBars()
            variable.Refresh(rect)
        dialog.Destroy()

    def EditConnectionContent(self, connection):
        dialog = ConnectionPropertiesDialog(self.ParentWindow)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetMinConnectionSize(connection.GetSize())
        values = {"name" : connection.GetName(), "type" : connection.GetType()}
        dialog.SetValues(values)
        if dialog.ShowModal() == wx.ID_OK:
            old_type = connection.GetType()
            values = dialog.GetValues()
            rect = connection.GetRedrawRect(1, 1)
            connection.SetName(values["name"])
            connection.SetType(values["type"])
            connection.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            rect = rect.Union(connection.GetRedrawRect())
            if old_type != values["type"]:
                id = connection.GetId()
                self.Controler.RemoveEditedElementInstance(self.TagName, id)
                self.Controler.AddEditedElementConnection(self.TagName, id, values["type"])
            self.RefreshConnectionModel(connection)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            connection.Refresh(rect)
        dialog.Destroy()

    def EditContactContent(self, contact):
        dialog = LDElementDialog(self.ParentWindow, "contact")
        dialog.SetPreviewFont(self.GetFont())
        varlist = []
        vars = self.Controler.GetEditedElementInterfaceVars(self.TagName)
        if vars:
            for var in vars:
                if var["Class"] != "Output" and var["Type"] == "BOOL":
                    varlist.append(var["Name"])
        dialog.SetVariables(varlist)
        values = {"name" : contact.GetName(), "type" : contact.GetType()}
        dialog.SetValues(values)
        dialog.SetElementSize(contact.GetSize())
        if dialog.ShowModal() == wx.ID_OK:
            values = dialog.GetValues()
            rect = contact.GetRedrawRect(1, 1)
            contact.SetName(values["name"])
            contact.SetType(values["type"])
            contact.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            rect = rect.Union(contact.GetRedrawRect())
            self.RefreshContactModel(contact)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            contact.Refresh(rect)
        dialog.Destroy()

    def EditCoilContent(self, coil):
        dialog = LDElementDialog(self.ParentWindow, "coil")
        dialog.SetPreviewFont(self.GetFont())
        varlist = []
        vars = self.Controler.GetEditedElementInterfaceVars(self.TagName)
        if vars:
            for var in vars:
                if var["Class"] != "Input" and var["Type"] == "BOOL":
                    varlist.append(var["Name"])
        returntype = self.Controler.GetEditedElementInterfaceReturnType(self.TagName)
        if returntype == "BOOL":
            varlist.append(self.Controler.GetEditedElementName(self.TagName))
        dialog.SetVariables(varlist)
        values = {"name" : coil.GetName(), "type" : coil.GetType()}
        dialog.SetValues(values)
        dialog.SetElementSize(coil.GetSize())
        if dialog.ShowModal() == wx.ID_OK:
            values = dialog.GetValues()
            rect = coil.GetRedrawRect(1, 1)
            coil.SetName(values["name"])
            coil.SetType(values["type"])
            coil.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            rect = rect.Union(coil.GetRedrawRect())
            self.RefreshCoilModel(coil)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            coil.Refresh(rect)
        dialog.Destroy()

    def EditPowerRailContent(self, powerrail):
        dialog = LDPowerRailDialog(self.ParentWindow, powerrail.GetType(), len(powerrail.GetConnectors()))
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetMinSize(powerrail.GetSize())
        if dialog.ShowModal() == wx.ID_OK:
            old_type = powerrail.GetType()
            values = dialog.GetValues()
            rect = powerrail.GetRedrawRect(1, 1)
            powerrail.SetType(values["type"], [True for i in xrange(values["number"])])
            powerrail.SetSize(*self.GetScaledSize(values["width"], values["height"]))
            rect = rect.Union(powerrail.GetRedrawRect())
            if old_type != values["type"]:
                id = powerrail.GetId()
                self.Controler.RemoveEditedElementInstance(self.TagName, id)
                self.Controler.AddEditedElementPowerRail(self.TagName, id, values["type"])
            self.RefreshPowerRailModel(powerrail)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            powerrail.Refresh(rect)
        dialog.Destroy()

    def EditStepContent(self, step):
        dialog = StepContentDialog(self.ParentWindow, step.GetInitial())
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetPouNames(self.Controler.GetProjectPouNames())
        dialog.SetVariables(self.Controler.GetEditedElementInterfaceVars(self.TagName))
        dialog.SetStepNames([block.GetName() for block in self.Blocks if isinstance(block, SFC_Step) and block.GetName() != step.GetName()])
        dialog.SetMinStepSize(step.GetSize())
        values = {"name" : step.GetName()}
        connectors = step.GetConnectors()
        values["input"] = connectors["input"] != None
        values["output"] = connectors["output"] != None
        values["action"] = connectors["action"] != None
        dialog.SetValues(values)
        if dialog.ShowModal() == wx.ID_OK:
            values = dialog.GetValues()
            rect = step.GetRedrawRect(1, 1)
            step.SetName(values["name"])
            if values["input"]:
                step.AddInput()
            else:
                step.RemoveInput()
            if values["output"]:
                step.AddOutput()
            else:
                step.RemoveOutput()
            if values["action"]:
                step.AddAction()    
            else:
                step.RemoveAction()
            step.UpdateSize(*self.GetScaledSize(values["width"], values["height"]))
            rect = rect.Union(step.GetRedrawRect())
            self.RefreshStepModel(step)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            step.Refresh(rect)
        
    def EditTransitionContent(self, transition):
        dialog = TransitionContentDialog(self.ParentWindow, self.GetDrawingMode() == FREEDRAWING_MODE)
        dialog.SetPreviewFont(self.GetFont())
        dialog.SetTransitions(self.Controler.GetEditedElementTransitions(self.TagName))
        dialog.SetValues({"type":transition.GetType(),"value":transition.GetCondition(), "priority":transition.GetPriority()})
        dialog.SetElementSize(transition.GetSize())
        if dialog.ShowModal() == wx.ID_OK:
            values = dialog.GetValues()
            rect = transition.GetRedrawRect(1, 1)
            transition.SetType(values["type"],values["value"])
            transition.SetPriority(values["priority"])
            rect = rect.Union(transition.GetRedrawRect())
            self.RefreshTransitionModel(transition)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            transition.Refresh(rect)
        dialog.Destroy()

    def EditJumpContent(self, jump):
        choices = []
        for block in self.Blocks:
            if isinstance(block, SFC_Step):
                choices.append(block.GetName())
        dialog = wx.SingleChoiceDialog(self.ParentWindow, "Edit jump target", "Please choose a target", choices, wx.OK|wx.CANCEL)
        dialog.SetSelection(choices.index(jump.GetTarget()))
        if dialog.ShowModal() == wx.ID_OK:
            value = dialog.GetStringSelection()
            rect = jump.GetRedrawRect(1, 1)
            jump.SetTarget(value)
            rect = rect.Union(jump.GetRedrawRect())
            self.RefreshJumpModel(jump)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            jump.Refresh(rect)
        dialog.Destroy()

    def EditActionBlockContent(self, actionblock):
        dialog = ActionBlockDialog(self.ParentWindow)
        dialog.SetQualifierList(self.Controler.GetQualifierTypes())
        dialog.SetActionList(self.Controler.GetEditedElementActions(self.TagName))
        dialog.SetVariableList(self.Controler.GetEditedElementInterfaceVars(self.TagName))
        dialog.SetValues(actionblock.GetActions())
        if dialog.ShowModal() == wx.ID_OK:
            actions = dialog.GetValues()
            rect = actionblock.GetRedrawRect(1, 1)
            actionblock.SetActions(actions)
            actionblock.SetSize(*self.GetScaledSize(*actionblock.GetSize()))
            rect = rect.Union(actionblock.GetRedrawRect())
            self.RefreshActionBlockModel(actionblock)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            actionblock.Refresh(rect)
        dialog.Destroy()

    def EditCommentContent(self, comment):
        if wx.VERSION >= (2, 5, 0):
            dialog = wx.TextEntryDialog(self.ParentWindow, "Edit comment", "Please enter comment text", comment.GetContent(), wx.OK|wx.CANCEL|wx.TE_MULTILINE)
        else:
            dialog = wx.TextEntryDialog(self.ParentWindow, "Edit comment", "Please enter comment text", comment.GetContent(), wx.OK|wx.CANCEL)
        if dialog.ShowModal() == wx.ID_OK:
            value = dialog.GetValue()
            rect = comment.GetRedrawRect(1, 1)
            comment.SetContent(value)
            comment.SetSize(*self.GetScaledSize(*comment.GetSize()))
            rect = rect.Union(comment.GetRedrawRect())
            self.RefreshCommentModel(comment)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            comment.Refresh(rect)
        dialog.Destroy()

#-------------------------------------------------------------------------------
#                          Model update functions
#-------------------------------------------------------------------------------

    def RefreshBlockModel(self, block):
        blockid = block.GetId()
        infos = {}
        infos["type"] = block.GetType()
        infos["name"] = block.GetName()
        if self.CurrentLanguage == "FBD":
            infos["executionOrder"] = block.GetExecutionOrder()
        infos["x"], infos["y"] = block.GetPosition()
        infos["width"], infos["height"] = block.GetSize()
        infos["connectors"] = block.GetConnectors()
        self.Controler.SetEditedElementBlockInfos(self.TagName, blockid, infos)
    
    def RefreshVariableModel(self, variable):
        variableid = variable.GetId()
        infos = {}
        infos["name"] = variable.GetName()
        if self.CurrentLanguage == "FBD":
            infos["executionOrder"] = variable.GetExecutionOrder()
        infos["x"], infos["y"] = variable.GetPosition()
        infos["width"], infos["height"] = variable.GetSize()
        infos["connectors"] = variable.GetConnectors()
        self.Controler.SetEditedElementVariableInfos(self.TagName, variableid, infos)

    def RefreshConnectionModel(self, connection):
        connectionid = connection.GetId()
        infos = {}
        infos["name"] = connection.GetName()
        infos["x"], infos["y"] = connection.GetPosition()
        infos["width"], infos["height"] = connection.GetSize()
        infos["connector"] = connection.GetConnector()
        self.Controler.SetEditedElementConnectionInfos(self.TagName, connectionid, infos)

    def RefreshCommentModel(self, comment):
        commentid = comment.GetId()
        infos = {}
        infos["content"] = comment.GetContent()
        infos["x"], infos["y"] = comment.GetPosition()
        infos["width"], infos["height"] = comment.GetSize()
        self.Controler.SetEditedElementCommentInfos(self.TagName, commentid, infos)

    def RefreshPowerRailModel(self, powerrail):
        powerrailid = powerrail.GetId()
        infos = {}
        infos["x"], infos["y"] = powerrail.GetPosition()
        infos["width"], infos["height"] = powerrail.GetSize()
        infos["connectors"] = powerrail.GetConnectors()
        self.Controler.SetEditedElementPowerRailInfos(self.TagName, powerrailid, infos)

    def RefreshContactModel(self, contact):
        contactid = contact.GetId()
        infos = {}
        infos["name"] = contact.GetName()
        infos["type"] = contact.GetType()
        infos["x"], infos["y"] = contact.GetPosition()
        infos["width"], infos["height"] = contact.GetSize()
        infos["connectors"] = contact.GetConnectors()
        self.Controler.SetEditedElementContactInfos(self.TagName, contactid, infos)

    def RefreshCoilModel(self, coil):
        coilid = coil.GetId()
        infos = {}
        infos["name"] = coil.GetName()
        infos["type"] = coil.GetType()
        infos["x"], infos["y"] = coil.GetPosition()
        infos["width"], infos["height"] = coil.GetSize()
        infos["connectors"] = coil.GetConnectors()
        self.Controler.SetEditedElementCoilInfos(self.TagName, coilid, infos)

    def RefreshStepModel(self, step):
        stepid = step.GetId()
        infos = {}
        infos["name"] = step.GetName()
        infos["initial"] = step.GetInitial()
        infos["x"], infos["y"] = step.GetPosition()
        infos["width"], infos["height"] = step.GetSize()
        infos["connectors"] = step.GetConnectors()
        self.Controler.SetEditedElementStepInfos(self.TagName, stepid, infos)

    def RefreshTransitionModel(self, transition):
        transitionid = transition.GetId()
        infos = {}
        infos["type"] = transition.GetType()
        infos["priority"] = transition.GetPriority()
        infos["condition"] = transition.GetCondition()
        infos["x"], infos["y"] = transition.GetPosition()
        infos["width"], infos["height"] = transition.GetSize()
        infos["connectors"] = transition.GetConnectors()
        self.Controler.SetEditedElementTransitionInfos(self.TagName, transitionid, infos)

    def RefreshDivergenceModel(self, divergence):
        divergenceid = divergence.GetId()
        infos = {}
        infos["x"], infos["y"] = divergence.GetPosition()
        infos["width"], infos["height"] = divergence.GetSize()
        infos["connectors"] = divergence.GetConnectors()
        self.Controler.SetEditedElementDivergenceInfos(self.TagName, divergenceid, infos)

    def RefreshJumpModel(self, jump):
        jumpid = jump.GetId()
        infos = {}
        infos["target"] = jump.GetTarget()
        infos["x"], infos["y"] = jump.GetPosition()
        infos["width"], infos["height"] = jump.GetSize()
        infos["connector"] = jump.GetConnector()
        self.Controler.SetEditedElementJumpInfos(self.TagName, jumpid, infos)

    def RefreshActionBlockModel(self, actionblock):
        actionblockid = actionblock.GetId()
        infos = {}
        infos["actions"] = actionblock.GetActions()
        infos["x"], infos["y"] = actionblock.GetPosition()
        infos["width"], infos["height"] = actionblock.GetSize()
        infos["connector"] = actionblock.GetConnector()
        self.Controler.SetEditedElementActionBlockInfos(self.TagName, actionblockid, infos)


#-------------------------------------------------------------------------------
#                          Model delete functions
#-------------------------------------------------------------------------------


    def DeleteBlock(self, block):
        elements = []
        for output in block.GetConnectors()["outputs"]:
            for element in output.GetConnectedBlocks():
                if element not in elements:
                    elements.append(element)
        block.Clean()
        self.RemoveBlock(block)
        self.Controler.RemoveEditedElementInstance(self.TagName, block.GetId())
        for element in elements:
            element.RefreshModel()
        wx.CallAfter(self.ParentWindow.RefreshVariablePanel, self.TagName)

    def DeleteVariable(self, variable):
        connectors = variable.GetConnectors()
        if connectors["output"]:
            elements = connectors["output"].GetConnectedBlocks()
        else:
            elements = []
        variable.Clean()
        self.RemoveBlock(variable)
        self.Controler.RemoveEditedElementInstance(self.TagName, variable.GetId())
        for element in elements:
            element.RefreshModel()

    def DeleteConnection(self, connection):
        if connection.GetType() == CONTINUATION:
            elements = connection.GetConnector().GetConnectedBlocks()
        else:
            elements = []
        connection.Clean()
        self.RemoveBlock(connection)
        self.Controler.RemoveEditedElementInstance(self.TagName, connection.GetId())
        for element in elements:
            element.RefreshModel()

    def DeleteComment(self, comment):
        self.RemoveComment(comment)
        self.Controler.RemoveEditedElementInstance(self.TagName, comment.GetId())

    def DeleteWire(self, wire):
        if wire in self.Wires:
            connected = wire.GetConnected()
            wire.Clean()
            self.RemoveWire(wire)
            for connector in connected:
                connector.RefreshParentBlock()

    def DeleteContact(self, contact):
        connectors = contact.GetConnectors()
        elements = connectors["output"].GetConnectedBlocks()
        contact.Clean()
        self.RemoveBlock(contact)
        self.Controler.RemoveEditedElementInstance(self.TagName, contact.GetId())
        for element in elements:
            element.RefreshModel()

    def DeleteCoil(self, coil):
        connectors = coil.GetConnectors()
        elements = connectors["output"].GetConnectedBlocks()
        coil.Clean()
        self.RemoveBlock(coil)
        self.Controler.RemoveEditedElementInstance(self.TagName, coil.GetId())
        for element in elements:
            element.RefreshModel()

    def DeletePowerRail(self, powerrail):
        elements = []
        if powerrail.GetType() == LEFTRAIL:
            for connector in powerrail.GetConnectors():
                for element in connector.GetConnectedBlocks():
                    if element not in elements:
                        elements.append(element)
        powerrail.Clean()
        self.RemoveBlock(powerrail)
        self.Controler.RemoveEditedElementInstance(self.TagName, powerrail.GetId())
        for element in elements:
            element.RefreshModel()

    def DeleteStep(self, step):
        elements = []
        connectors = step.GetConnectors()
        if connectors["output"]:
            for element in connectors["output"].GetConnectedBlocks():
                if element not in elements:
                    elements.append(element)
        if connectors["action"]:
            for element in connectors["action"].GetConnectedBlocks():
                if element not in elements:
                    elements.append(element)
        step.Clean()
        self.RemoveBlock(step)
        self.Controler.RemoveEditedElementInstance(self.TagName, step.GetId())
        for element in elements:
            element.RefreshModel()
            
    def DeleteTransition(self, transition):
        elements = []
        connectors = transition.GetConnectors()
        if connectors["output"]:
            for element in connectors["output"].GetConnectedBlocks():
                if element not in elements:
                    elements.append(element)
        transition.Clean()
        self.RemoveBlock(transition)
        self.Controler.RemoveEditedElementInstance(self.TagName, transition.GetId())
        for element in elements:
            element.RefreshModel()

    def DeleteDivergence(self, divergence):
        elements = []
        connectors = divergence.GetConnectors()
        for output in connectors["outputs"]:
            for element in output.GetConnectedBlocks():
                if element not in elements:
                    elements.append(element)
        divergence.Clean()
        self.RemoveBlock(divergence)
        self.Controler.RemoveEditedElementInstance(self.TagName, divergence.GetId())
        for element in elements:
            element.RefreshModel()
    
    def DeleteJump(self, jump):
        jump.Clean()
        self.RemoveBlock(jump)
        self.Controler.RemoveEditedElementInstance(self.TagName, jump.GetId())
    
    def DeleteActionBlock(self, actionblock):
        actionblock.Clean()
        self.RemoveBlock(actionblock)
        self.Controler.RemoveEditedElementInstance(self.TagName, actionblock.GetId())


#-------------------------------------------------------------------------------
#                            Editing functions
#-------------------------------------------------------------------------------
    
    def Cut(self):
        if self.IsBlock(self.SelectedElement) or self.IsComment(self.SelectedElement):
            self.ParentWindow.SetCopyBuffer(self.SelectedElement.Clone(self))
            rect = self.SelectedElement.GetRedrawRect(1, 1)
            self.SelectedElement.Delete()
            self.SelectedElement = None
            self.RefreshBuffer()
            self.RefreshScrollBars()
            self.RefreshRect(self.GetScrolledRect(rect), False)
        
    def Copy(self):
        if self.IsBlock(self.SelectedElement) or self.IsComment(self.SelectedElement):
            self.ParentWindow.SetCopyBuffer(self.SelectedElement.Clone(self))
            
    def Paste(self):
        element = self.ParentWindow.GetCopyBuffer()
        if element is not None and self.CanAddBlock(element):
            block = self.CopyBlock(element, wx.Point(*self.CalcUnscrolledPosition(30, 30)))
            if self.SelectedElement is not None:
                self.SelectedElement.SetSelected(False)
            self.SelectedElement = block
            self.SelectedElement.SetSelected(True)
            self.RefreshBuffer()
            self.RefreshScrollBars()
            self.ParentWindow.RefreshVariablePanel(self.TagName)

    def CanAddBlock(self, block):
        if self.CurrentLanguage == "SFC":
            return True
        elif self.CurrentLanguage == "LD" and not isinstance(block, (SFC_Step, SFC_Transition, SFC_Divergence, SFC_Jump, SFC_ActionBlock)):
            return True
        elif self.CurrentLanguage == "FBD" and isinstance(block, (FBD_Block, FBD_Variable, FBD_Connector, Comment)):
            return True
        return False

    def CopyBlock(self, element, pos):
        id = self.GetNewId()
        if isinstance(element, FBD_Block) and element.GetName() != "" or isinstance(element, SFC_Step):
            if isinstance(element, FBD_Block):
                names = [varname.upper() for varname in self.Controler.GetEditedElementVariables(self.TagName)]
                format = "Block%d"
            elif isinstance(element, SFC_Step):
                names = [block.GetName().upper() for block in self.Blocks if isinstance(block, SFC_Step)]
                format = "Step%d"
            i = 1
            while (format%i).upper() in names:
                i += 1
            name = format%i
            block = element.Clone(self, id, name, pos)
        else:
            name = None
            block = element.Clone(self, id, pos=pos)
        if isinstance(block, Comment):
            self.AddComment(block)
            self.Controler.AddEditedElementComment(self.TagName, id)
            self.RefreshCommentModel(block)
        else:
            self.AddBlock(block)
            if isinstance(block, FBD_Block):
                self.Controler.AddEditedElementBlock(self.TagName, id, block.GetType(), name)
                self.RefreshBlockModel(block)
            elif isinstance(block, FBD_Variable):
                self.Controler.AddEditedElementVariable(self.TagName, id, block.GetType())
                self.RefreshVariableModel(block)
            elif isinstance(block, FBD_Connector):
                self.Controler.AddEditedElementConnection(self.TagName, id, block.GetType())
                self.RefreshConnectionModel(block)
            elif isinstance(block, LD_Contact):
                self.Controler.AddEditedElementContact(self.TagName, id)
                self.RefreshContactModel(block)
            elif isinstance(block, LD_Coil):
                self.Controler.AddEditedElementCoil(self.TagName, id)
                self.RefreshCoilModel(block)
            elif isinstance(block, LD_PowerRail):
                self.Controler.AddEditedElementPowerRail(self.TagName, id, block.GetType())
                self.RefreshPowerRailModel(block)
            elif isinstance(block, SFC_Step):
                self.Controler.AddEditedElementStep(self.TagName, id)
                self.RefreshStepModel(block)    
            elif isinstance(block, SFC_Transition):
                self.Controler.AddEditedElementTransition(self.TagName, id)
                self.RefreshTransitionModel(block)       
            elif isinstance(block, SFC_Divergence):
                self.Controler.AddEditedElementDivergence(self.TagName, id, block.GetType())
                self.RefreshDivergenceModel(block)
            elif isinstance(block, SFC_Jump):
                self.Controler.AddEditedElementJump(self.TagName, id)
                self.RefreshJumpModel(block)       
            elif isinstance(block, SFC_ActionBlock):
                self.Controler.AddEditedElementActionBlock(self.TagName, id)
                self.RefreshActionBlockModel(block)
        return block
            

#-------------------------------------------------------------------------------
#                            Drawing functions
#-------------------------------------------------------------------------------

    def OnMoveWindow(self, event):
        self.GetBestSize()
        self.RefreshScrollBars()
        event.Skip()

    def OnPaint(self, event):
        dc = self.GetLogicalDC(True)
        dc.SetBackground(wx.Brush(self.GetBackgroundColour()))
        dc.Clear()
        dc.BeginDrawing()
        if self.Scaling and self.DrawGrid:
            dc.SetPen(wx.TRANSPARENT_PEN)
            dc.SetBrush(self.GridBrush)
            width, height = self.GetVirtualSize()
            dc.DrawRectangle(0, 0, width, height)
        
        # Draw all elements
        for comment in self.Comments:
            if comment != self.SelectedElement:
                comment.Draw(dc)
        for wire in self.Wires:
            if wire != self.SelectedElement:
                wire.Draw(dc)
        for block in self.Blocks:
            if block != self.SelectedElement:
                block.Draw(dc)
        
        if self.SelectedElement:
            self.SelectedElement.Draw(dc)
        
        if self.rubberBand.IsShown():
            self.rubberBand.Draw(dc)
        dc.EndDrawing()
        event.Skip()


